"""Interaction energy functionals.

Implements field--particle interaction energy functionals depending on the
field configuration. Extending HyMD with a new Hamiltonian is done by
subclassing the :code:`Hamiltonian` superclass with logic intialized in the
constructor.

The grid-independent filtering function is defined in the :code:`Hamiltonian`
superclass constructor. Hijack :code:`self.H` in a subclass to change the
filter.
"""
import numpy as np
import sympy


class Hamiltonian:
    """Interaction energy functional superclass
    """
    def __init__(self, config):
        """Constructor

        Parameters
        ----------
        config : Config
            Configuration object.

        See also
        --------
        hymd.input_parser.Config : Configuration dataclass handler.
        """
        self.config = config
        self._setup()

    def _setup(self):
        """Superclass setup

        Sets up the grid-independent filtering function :code:`H`, and the
        SymPy logic for symbolically differentiating interaction energy
        functionals in Hamiltonian subclasses.
        """
        if not hasattr(self.config, "simulation_volume"):
            self.config.simulation_volume = np.prod(
                np.asarray(self.config.box_size)
            )
        self.config.rho0 = (
            self.config.n_particles / self.config.simulation_volume
        )
        self.phi = sympy.var("phi:%d" % (len(self.config.unique_names)))
        k = sympy.var("k:%d" % (3))

        def fourier_space_window_function(k):
            return sympy.functions.elementary.exponential.exp(
                -0.5
                * self.config.sigma ** 2
                * (k0 ** 2 + k1 ** 2 + k2 ** 2)  # noqa: F821, E501
            )

        self.window_function_lambda = sympy.lambdify(
            [k], fourier_space_window_function(k)
        )

        def H(k, v):
            return v * self.window_function_lambda(k)

        self.H = H


class SquaredPhi(Hamiltonian):
    """Simple squared density interaction energy functional

    The interaction energy density takes the form

    .. math::

        w[\\tilde\\phi] = \\frac{1}{2\\kappa\\rho_0}
            \\left(
                \\sum_k \\tilde\\phi_k
            \\right)^2,

    where :math:`\\kappa` is the compressibility and :math:`rho_0` is the
    average density of the fully homogenous system. Expressing the species
    densities in terms of fluctuations from the average,

    .. math::

        \\mathrm{d}\\tilde\\phi_k = \\rho_0 - \\tilde\\phi_k,

    it is evident that this interaction energy functional is a slight change of
    the :code:`DefaultNoChi` Hamiltonian, as the expanded interaction energy
    density becomes

    .. math::

        w[\\tilde\\phi] = \\frac{1}{2\\kappa\\rho_0}
            \\left(
                6\\rho_0\\left[
                    \\sum_k \\mathrm{d}\\tilde\\phi_k
                \\right]
                +
                9\\rho_0^2
                +
                \\sum_k \\mathrm{d}\\tilde\\phi_k^2
                +
                \\prod_{k\\not=l}
                    \\mathrm{d}\\tilde\\phi_k \\mathrm{d}\\tilde\\phi_l
            \\right),

    identical to :code:`DefaultNoChi` apart from a constant energy shift
    :math:`9\\rho_0^2`(which does not impact dynamics) and the
    :math:`\\rho_0`--:math:`\\mathrm{d}\\tilde\\phi_k` cross-terms. These
    cross-terms constitute contributions to the energy linear in
    :math:`\\mathrm{d}\\tilde\\phi_k` absent in :code:`DefaultNoChi`, which has
    only quadratic terms present.

    See also
    --------
    hymd.hamiltonian.DefaultNoChi
    """
    def __init__(self, config):
        """Constructor

        Parameters
        ----------
        config : Config
            Configuration object.

        See also
        --------
        hymd.input_parser.Config : Configuration dataclass handler.
        """
        super(SquaredPhi, self).__init__(config)
        super(SquaredPhi, self)._setup()
        self.setup()

    def setup(self):
        """Setup the interaction energy potential and the external potential
        """
        def w(phi, kappa=self.config.kappa, rho0=self.config.rho0):
            return 0.5 / (kappa * rho0) * (sum(phi)) ** 2

        self.v_ext = [
            sympy.lambdify([self.phi], sympy.diff(w(self.phi), self.phi[i]))
            for i in range(len(self.config.unique_names))
        ]
        self.w = sympy.lambdify([self.phi], w(self.phi))


class DefaultNoChi(Hamiltonian):
    """Incompressibility-only interaction energy functional

    The interaction energy density takes the form

    .. math::

        w[\\tilde\\phi] = \\frac{1}{2\\kappa} \\left(
            \\sum_k \\tilde\\phi_k - \\rho_0
        \\right)^2,

    where :math:`\\kappa` is the compressibility and :math:`\\rho_0` is the
    average density of the fully homogenous system. The :code:`SquaredPhi`
    Hamiltonian implements a similar functional with an additional linear term
    component depending on

    .. math::

        \\mathmr{d}\\tilde\\phi_k = \\tilde\\phi_k - \\rho_0

    and not :math:`\\mathrm{d}\\tilde\\phi_k^2`. No explicit inter-species
    interaction is present apart from the indirect interaction through the
    incompressibility.

    See also
    --------
    hymd.hamiltonian.DefaultNoChi
    hymd.input_parser.Config :
        Configuration dataclass handler.
    """
    def __init__(self, config):
        """Constructor

        Parameters
        ----------
        config : Config
            Configuration object.

        See also
        --------
        hymd.input_parser.Config : Configuration dataclass handler.
        """
        super(DefaultNoChi, self).__init__(config)
        super(DefaultNoChi, self)._setup()
        self.setup()

    def setup(self):
        """Setup the interaction energy potential and the external potential
        """
        def w(phi, kappa=self.config.kappa, rho0=self.config.rho0):
            return 0.5 / (kappa * rho0) * (sum(phi) - rho0) ** 2

        self.v_ext = [
            sympy.lambdify([self.phi], sympy.diff(w(self.phi), self.phi[i]))
            for i in range(len(self.config.unique_names))
        ]
        self.w = sympy.lambdify([self.phi], w(self.phi))


class DefaultWithChi(Hamiltonian):
    """Incompressibility and :math:`\\chi`-interactions energy functional

    The interaction energy density takes the form

    .. math::

        w[\\tilde\\phi] =
            \\frac{1}{2\\rho_0}
                \\sum_{k,l}\\chi_{kl} \\tilde\\phi_k \\tilde\\phi_l
            +
            \\frac{1}{2\\kappa} \\left(
                \\sum_k \\tilde\\phi_k - \\rho_0
            \\right)^2,

    where :math:`\\kappa` is the incompressibility, :math:`\\rho_0` is the
    average density of the fully homogenous system and :math:`\\chi_{ij}` is
    the Flory-Huggins-like inter-species mixing energy.
    """
    def __init__(self, config, unique_names, type_to_name_map):
        """Constructor

        Parameters
        ----------
        config : Config
            Configuration object.
        unique_names : numpy.ndarray
            Sorted array of all distinct names of different species present in
            the simulation. Result of :code:`numpy.unique(all_names)`, where
            :code:`all_names` is the gathered union of all individual MPI
            ranks' :code:`names` arrays.
        type_to_name_map : dict[int, str]
            Dictionary of the mapping from type indices (integers) to type
            names.

        See also
        --------
        hymd.input_parser.Config : Configuration dataclass handler.
        """
        super(DefaultWithChi, self).__init__(config)
        super(DefaultWithChi, self)._setup()
        self.setup(unique_names, type_to_name_map)

    def setup(self, unique_names, type_to_name_map):
        """Setup the interaction energy potential and the external potential

        Parameters
        ----------
        unique_names : numpy.ndarray
            Sorted array of all distinct names of different species present in
            the simulation. Result of :code:`numpy.unique(all_names)`, where
            :code:`all_names` is the gathered union of all individual MPI
            ranks' :code:`names` arrays.
        type_to_name_map : dict[int, str]
            Dictionary of the mapping from type indices (integers) to type
            names.
        """
        self.type_to_name_map = type_to_name_map
        self.chi_type_dictionary = {
            tuple(sorted([c.atom_1, c.atom_2])): c.interaction_energy
            for c in self.config.chi
        }

        def w(
            phi,
            kappa=self.config.kappa,
            rho0=self.config.rho0,
            chi=self.config.chi,
            type_to_name_map=self.type_to_name_map,
            chi_type_dictionary=self.chi_type_dictionary,
        ):

            interaction = 0
            for i in range(self.config.n_types):
                for j in range(i + 1, self.config.n_types):
                    ni = type_to_name_map[i]
                    nj = type_to_name_map[j]
                    names = sorted([ni, nj])
                    c = chi_type_dictionary[tuple(names)]

                    interaction += c * phi[i] * phi[j] / rho0
            incompressibility = 0.5 / (kappa * rho0) * (sum(phi) - rho0) ** 2
            return incompressibility + interaction

        self.v_ext = [
            sympy.lambdify([self.phi], sympy.diff(w(self.phi), self.phi[i]))
            for i in range(len(self.config.unique_names))
        ]
        self.w = sympy.lambdify([self.phi], w(self.phi))



"""Forces and energies from the discretized particle-field grid interactions
"""
import logging
import numpy as np
from mpi4py import MPI
from .logger import Logger


def compute_field_force(layouts, r, force_mesh, force, types, n_types):
    """Interpolate particle-field forces from the grid onto particle positions

    Backmaps the forces calculated on the grid to particle positions using the
    window function :math:`P` (by default cloud-in-cell [CIC]). In the
    following, let :math:`\\mathbf{F}_j` denote the force acting on the
    particle with index :math:`j` and position :math:`\\mathbf{r}_j`. The
    interpolated force is

    .. math::

        \\mathbf{F}_j = -\\sum_k\\nabla V_{j_k}
            P(\\mathbf{r}_{j_k}-\\mathbf{r}_j)h^3,

    where :math:`V_{j_k}` is the discretized external potential at grid vertex
    :math:`j_k`, :math:`\\mathbf{r}_{j_k}` is the position of the grid vertex
    with index :math:`j_k`, and :math:`h^3` is the volume of each grid voxel.
    The sum is taken over all closest neighbour vertices, :math:`j_k`.

    Parameters
    ----------
    layouts : list[pmesh.domain.Layout]
        Pmesh communication layout objects for domain decompositions of each
        particle type. Used as blueprint by :code:`pmesh.pm.readout` for
        exchange of particle information across MPI ranks as necessary.
    r : (N,D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    force_mesh : list[pmesh.pm.RealField]
        Pmesh :code:`RealField` objects containing discretized particle-field
        force density values on the computational grid; :code:`D` fields in D
        dimensions for each particle type. Local for each MPI rank--the full
        computaional grid is represented by the collective fields of all MPI
        ranks.
    force : (N,D) numpy.ndarray
        Array of forces for :code:`N` particles in :code:`D` dimensions. Local
        for each MPI rank.
    types : (N,) numpy.ndarray
        Array of type indices for each of :code:`N` particles. Local for each
        MPI rank.
    n_types : int
        Number of different unique types present in the simulation system.
        :code:`n_types` is global, i.e. the same for all MPI ranks even if some
        ranks contain zero particles of certain types.
    """
    for t in range(n_types):
        ind = types == t
        for d in range(3):
            force[ind, d] = force_mesh[t][d].readout(r[ind], layout=layouts[t])


def compute_field_energy_q(
    config, phi_q_fourier, elec_energy_field, field_q_energy,
    comm=MPI.COMM_WORLD,
):
    """Calculate the electrostatic energy from a field configuration

    From the definition of the elecrostatic potential :math:`\\Psi`, the energy
    is

    .. math::

        E = \\frac{1}{2}\\int\\mathrm{d}\\mathbf{r}\\,
            \\rho(\\mathbf{r}) \\Psi(\\mathbf{r}),

    where :math:`\\rho(\\mathbf{r})` denotes the charge density at position
    :math:`\\mathbf{r}`. Through application of Gauss' law and writing
    :math:`\\Psi` in terms of the electric field :math:`\\mathbf{E}`, this
    becomes

    .. math::

        E = \\frac{\\varepsilon}{2}\\int\\mathrm{d}\\mathbf{r}\\,
            \\mathbf{E}\\cdot \\mathbf{E},

    where :math:`\\varepsilon` is the relative dielectric of the simulation
    medium.

    Parameters
    ----------
    config : Config
        Configuration object.
    phi_q_fourier : pmesh.pm.ComplexField
        Pmesh :code:`ComplexField` object containing the discretized
        electrostatic potential values in reciprocal space on the
        computational grid. Local for each MPI rank--the full computaional grid
        is represented by the collective fields of all MPI ranks.
    elec_energy_field : pmesh.pm.ComplexField
        Pmesh :code:`ComplexField` object for storing calculated discretized
        electrostatic energy density values in reciprocal space on the
        computational grid. Pre-allocated, but empty; any values in this field
        are discarded. Changed in-place. Local for each MPI rank--the full
        computaional grid is represented by the collective fields of all MPI
        ranks.
    field_q_energy : float
        Total elecrostatic energy.

    Returns
    -------
    elec_field_energy : float
        Total electrostatic energy.
    """
    elec_conversion_factor = config.coulomb_constant / config.dielectric_const

    def transfer_energy(k, v):
        return (
            4.0 * np.pi * elec_conversion_factor * np.abs(v)**2
            / k.normp(p=2, zeromode=1)
        )

    phi_q_fourier.apply(
        transfer_energy, kind="wavenumber", out=elec_energy_field
    )
    V = np.prod(config.box_size)
    field_q_energy = 0.5 * V * comm.allreduce(np.sum(elec_energy_field.value))
    return field_q_energy.real


def update_field_force_q(
    charges, phi_q, phi_q_fourier, elec_field_fourier, elec_field,
    elec_forces, layout_q, pm, positions, config,
):
    """Calculate the electrostatic particle-field forces on the grid

    Computes the electrostatic potential :math:`\\Psi` from particle charges
    through the smoothed charge density :math:`\\tilde\\rho`. With :math:`P`
    being the cloud-in-cell (CIC) window function, the charge density and
    filtered charge densities are computed as

    .. math::

        \\rho(\\mathbf{r}) = \\sum_i q_i P(\\mathbf{r}-\\mathbf{r}_i),

    and

    .. math::

        \\tilde\\rho(\\mathbf{r}) = \\int\\mathrm{x}\\mathbf{r}\\,
            \\rho(\\mathbf{x})H(\\mathbf{r}-\\mathbf{x}),

    where :math:`H` is the grid-independent filtering function. The
    electrostatic potential is computed in reciprocal space as

    .. math::

        \\Phi = \\mathrm{FFT}^{-1}\\left[
            \\frac{4\\pi k_e}{\\varepsilon \\vert\\mathbf{k}\\vert^2}
            \\mathrm{FFT}(\\rho)\\mathrm{FFT}(H)
        \\right],

    with the electric field

    .. math::

        \\mathbf{E} = \\mathrm{FFT}^{-1}\\left[
            -i\\mathbf{k}\\,\\mathrm{FFT}(\\Psi)
        \\right].

    In the following, let :math:`\\mathbf{F}_j` denote the electrostatic force
    acting on the particle with index :math:`j` and position
    :math:`\\mathbf{r}_j`. The interpolated electrostatic force is

    .. math::

        \\mathbf{F}_j = \\sum_k q_j\\mathbf{E}_{j_k}
            P(\\mathbf{r}_{j_k}-\\mathbf{r}_j)h^3,

    where :math:`\\mathbf{E}_{j_k}` is the discretized electric field at grid
    vertex :math:`j_k`, :math:`\\mathbf{r}_{j_k}` is the position of the grid
    vertex with index :math:`j_k`, and :math:`h^3` is the volume of each grid
    voxel. The sum is taken over all closest neighbour vertices, :math:`j_k`.

    Parameters
    ----------
    charges : (N,) numpy.ndarray
        Array of particle charge values for :code:`N` particles. Local for each
        MPI rank.
    phi_q : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing calculated discretized
        charge density density values on the computational grid. Pre-allocated,
        but empty; any values in this field are discarded. Changed in-place.
        Local for each MPI rank--the full computaional grid is represented by
        the collective fields of all MPI ranks.
    phi_q_fourier : pmesh.pm.ComplexField
        Pmesh :code:`ComplexField` object for storing calculated discretized
        Fourier transformed charge density values in reciprocal space on the
        computational grid. Pre-allocated, but empty; any values in this field
        are discarded. Changed in-place. Local for each MPI rank--the full
        computaional grid is represented by the collective fields of all MPI
        ranks.
    elec_field_fourier : pmesh.pm.ComplexField
        Pmesh :code:`ComplexField` object for storing calculated discretized
        electric field values in reciprocal space on the computational grid.
        Pre-allocated, but empty; any values in this field are discarded.
        Changed in-place. Local for each MPI rank--the full computaional grid
        is represented by the collective fields of all MPI ranks.
    elec_field : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing calculated discretized
        electric field values on the computational grid. Pre-allocated,
        but empty; any values in this field are discarded. Changed in-place.
        Local for each MPI rank--the full computaional grid is represented by
        the collective fields of all MPI ranks.
    elec_forces : (N,D) numpy.ndarray
        Array of electrostatic forces on :code:`N` particles in :code:`D`
        dimensions.
    layout_q : pmesh.domain.Layout
        Pmesh communication layout object for domain decomposition of the full
        system. Used as blueprint by :code:`pmesh.pm.paint` and
        :code:`pmesh.pm.readout` for exchange of particle information across
        MPI ranks as necessary.
    pm : pmesh.pm.ParticleMesh
        Pmesh :code:`ParticleMesh` object interfacing to the CIC window
        function and the PFFT discrete Fourier transform library.
    positions : (N,D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    config : hymd.input_parser.Config
        Configuration object.
    """
    V = np.prod(config.box_size)
    n_mesh_cells = np.prod(np.full(3, config.mesh_size))
    volume_per_cell = V / n_mesh_cells
    pm.paint(positions, layout=layout_q, mass=charges, out=phi_q)
    phi_q /= volume_per_cell
    phi_q.r2c(out=phi_q_fourier)

    def phi_transfer_function(k, v):
        return v * np.exp(-0.5 * config.sigma**2 * k.normp(p=2, zeromode=1))

    phi_q_fourier.apply(phi_transfer_function, out=phi_q_fourier)
    phi_q_fourier.c2r(out=phi_q)
    n_dimensions = config.box_size.size
    elec_conversion_factor = config.coulomb_constant / config.dielectric_const

    for _d in np.arange(n_dimensions):

        def poisson_transfer_function(k, v, d=_d):
            return (
                -1j * k[d] * 4.0 * np.pi * elec_conversion_factor * v
                / k.normp(p=2, zeromode=1)
            )
        phi_q_fourier.apply(
            poisson_transfer_function, out=elec_field_fourier[_d]
        )
        elec_field_fourier[_d].c2r(out=elec_field[_d])

    for _d in np.arange(n_dimensions):
        elec_forces[:, _d] = charges * (
            elec_field[_d].readout(positions, layout=layout_q)
        )


def update_field_force_energy_q(
    charges, phi_q, phi_q_fourier, elec_field_fourier, elec_field, elec_forces,
    elec_energy_field, field_q_energy, layout_q, pm, positions, config,
    compute_energy=False, comm=MPI.COMM_WORLD,
):
    """Calculate the electrostatic particle-field energy and forces on the grid

    .. deprecated:: 1.0.0
        :code:`update_field_force_energy_q` was deprecated in favour of
        :code:`update_field_force_q` and independent calls to
        :code:`compute_field_energy_q` prior to 1.0.0 release.

    Parameters
    ----------
    charges : (N,) numpy.ndarray
        Array of particle charge values for :code:`N` particles. Local for each
        MPI rank.
    phi_q : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing calculated discretized
        charge density density values on the computational grid. Pre-allocated,
        but empty; any values in this field are discarded. Changed in-place.
        Local for each MPI rank--the full computaional grid is represented by
        the collective fields of all MPI ranks.
    phi_q_fourier : pmesh.pm.ComplexField
        Pmesh :code:`ComplexField` object for storing calculated discretized
        Fourier transformed charge density values in reciprocal space on the
        computational grid. Pre-allocated, but empty; any values in this field
        are discarded. Changed in-place. Local for each MPI rank--the full
        computaional grid is represented by the collective fields of all MPI
        ranks.
    elec_field_fourier : pmesh.pm.ComplexField
        Pmesh :code:`ComplexField` object for storing calculated discretized
        electric field values in reciprocal space on the computational grid.
        Pre-allocated, but empty; any values in this field are discarded.
        Changed in-place. Local for each MPI rank--the full computaional grid
        is represented by the collective fields of all MPI ranks.
    elec_field : pmesh.pm.RealField
        Pmesh :code:`RealField` object for storing calculated discretized
        electric field values on the computational grid. Pre-allocated,
        but empty; any values in this field are discarded. Changed in-place.
        Local for each MPI rank--the full computaional grid is represented by
        the collective fields of all MPI ranks.
    elec_forces : (N,D) numpy.ndarray
        Array of electrostatic forces on :code:`N` particles in :code:`D`
        dimensions.
    elec_energy_field : pmesh.pm.ComplexField
        Pmesh :code:`ComplexField` object for storing calculated discretized
        electrostatic energy density values in reciprocal space on the
        computational grid. Pre-allocated, but empty; any values in this field
        are discarded. Changed in-place. Local for each MPI rank--the full
        computaional grid is represented by the collective fields of all MPI
        ranks.
    field_q_energy : float
        Total elecrostatic energy.
    layout_q : pmesh.domain.Layout
        Pmesh communication layout object for domain decomposition of the full
        system. Used as blueprint by :code:`pmesh.pm.paint` and
        :code:`pmesh.pm.readout` for exchange of particle information across
        MPI ranks as necessary.
    pm : pmesh.pm.ParticleMesh
        Pmesh :code:`ParticleMesh` object interfacing to the CIC window
        function and the PFFT discrete Fourier transform library.
    positions : (N,D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    config : Config
        Configuration object.
    compute_energy : bool, optional
        Computes the electrostatic energy if :code:`True`, otherwise only
        computes the electrostatic forces.
    comm : mpi4py.Comm
        MPI communicator to use for rank commuication.

    Returns
    -------
    elec_field_energy : float
        Total electrostatic energy.
    """
    V = np.prod(config.box_size)
    n_mesh_cells = np.prod(np.full(3, config.mesh_size))
    volume_per_cell = V / n_mesh_cells
    pm.paint(positions, layout=layout_q, mass=charges, out=phi_q)
    phi_q /= volume_per_cell
    phi_q.r2c(out=phi_q_fourier)

    def phi_transfer_function(k, v):
        return v * np.exp(-0.5 * config.sigma ** 2 * k.normp(p=2, zeromode=1))

    phi_q_fourier.apply(phi_transfer_function, out=phi_q_fourier)
    phi_q_fourier.c2r(out=phi_q)
    n_dimensions = config.box_size.size
    elec_conversion_factor = config.coulomb_constant / config.dielectric_const

    for _d in np.arange(n_dimensions):
        def poisson_transfer_function(k, v, d=_d):
            return (
                -1j * k[d] * 4.0 * np.pi * elec_conversion_factor * v
                / k.normp(p=2, zeromode=1)
            )

        phi_q_fourier.apply(
            poisson_transfer_function, out=elec_field_fourier[_d]
        )
        elec_field_fourier[_d].c2r(out=elec_field[_d])

    for _d in np.arange(n_dimensions):
        elec_forces[:, _d] = charges * (
            elec_field[_d].readout(positions, layout=layout_q)
        )

    if compute_energy:
        def transfer_energy(k, v):
            return (
                4.0 * np.pi * elec_conversion_factor * np.abs(v)**2
                / k.normp(p=2, zeromode=1)
            )
        phi_q_fourier.apply(
            transfer_energy, kind="wavenumber", out=elec_energy_field
        )
        field_q_energy = 0.5 * comm.allreduce(np.sum(elec_energy_field.value))

    return field_q_energy.real


def update_field(
    phi, layouts, force_mesh, hamiltonian, pm, positions, types, config, v_ext,
    phi_fourier, v_ext_fourier, compute_potential=False,
):
    """Calculate the particle-field potential and force density

    If :code:`compute_potential` is :code:`True`, the energy may subsequently
    be computed by calling :code:`compute_field_and_kinetic_energy`.

    Computes the particle-field external potential :math:`V_\\text{ext}` from
    particle number densities through the smoothed density field,
    :math:`\\phi(\\mathbf{r})`. With :math:`P` being the cloud-in-cell (CIC)
    window function, the density and filtered densities are computed as

    .. math::

        \\phi(\\mathbf{r}) = \\sum_i P(\\mathbf{r}-\\mathbf{r}_i),

    and

    .. math::

        \\tilde\\phi(\\mathbf{r}) = \\int\\mathrm{x}\\mathbf{r}\\,
            \\phi(\\mathbf{x})H(\\mathbf{r}-\\mathbf{x}),

    where :math:`H` is the grid-independent filtering function. The
    external potential is computed in reciprocal space as

    .. math::

        V_\\text{ext} = \\mathrm{FFT}^{-1}\\left[
            \\mathrm{FFT}
                \\left(
                    \\frac{\\partial w}{\\partial \\tilde\\phi}
                \\right)
            \\mathrm{FFT}(H)
        \\right],

    where :math:`w` is the interaction energy functional. Differentiating
    :math:`V_\\text{ext}` is done by simply applying :math:`i\\mathbf{k}` in
    Fourier space, and the resulting forces are back-transformed to direct
    space and interpolated to particle positions by

    .. math::

        \\mathbf{F}_j = -\\sum_{j_k} \\nabla V_{j_k}
            P(\\mathbf{r}_{j_k} - \\mathbf{r}_j)h^3,

    where :math:`j_k` are the neighbouring vertices of particle :math:`j` at
    position :math:`\\mathbf{r}_j`, and :math:`h^3` is the volume of each grid
    voxel.

    Parameters
    ----------
    phi : list[pmesh.pm.RealField]
        Pmesh :code:`RealField` objects containing discretized particle number
        density values on the computational grid; one for each particle type.
        Pre-allocated, but empty; any values in this field are discarded.
        Changed in-place. Local for each MPI rank--the full computaional grid
        is represented by the collective fields of all MPI ranks.
    layouts : list[pmesh.domain.Layout]
        Pmesh communication layout objects for domain decompositions of each
        particle type. Used as blueprint by :code:`pmesh.pm.readout` for
        exchange of particle information across MPI ranks as necessary.
    force_mesh : list[pmesh.pm.RealField]
        Pmesh :code:`RealField` objects containing discretized particle-field
        force density values on the computational grid; :code:`D` fields in D
        dimensions for each particle type. Pre-allocated, but empty; any values
        in this field are discarded. Changed in-place. Local for each MPI
        rank--the full computaional grid is represented by the collective
        fields of all MPI ranks.
    hamiltonian : Hamiltonian
        Particle-field interaction energy handler object. Defines the
        grid-independent filtering function, :math:`H`.
    pm : pmesh.pm.ParticleMesh
        Pmesh :code:`ParticleMesh` object interfacing to the CIC window
        function and the PFFT discrete Fourier transform library.
    positions : (N,D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    types : (N,) numpy.ndarray
        Array of type indices for each of :code:`N` particles. Local for each
        MPI rank.
    config : Config
        Configuration object.
    v_ext : list[pmesh.pm.RealField]
        Pmesh :code:`RealField` objects containing discretized particle-field
        external potential values on the computational grid; one for each
        particle type. Pre-allocated, but empty; any values in this field are
        discarded Changed in-place. Local for each MPI rank--the full
        computaional grid is represented by the collective fields of all MPI
        ranks.
    phi_fourier : list[pmesh.pm.ComplexField]
        Pmesh :code:`ComplexField` objects containing discretized particle
        number density values in reciprocal space on the computational grid;
        one for each particle type. Pre-allocated, but empty; any values in
        this field are discarded Changed in-place. Local for each MPI rank--the
        full computaional grid is represented by the collective fields of all
        MPI ranks.
    v_ext_fourier : list[pmesh.pm.ComplexField]
        Pmesh :code:`ComplesField` objects containing discretized
        particle-field external potential values in reciprocal space on the
        computational grid; :code:`D+1` fields in D dimensions for each
        particle type. :code:`D` copies are made after calculation for later
        use in force calculation, because the `force transfer function`
        application differentiates the field in-place, ruining the contents
        for differentiation in the remaining :code:`D-1` spatial directions.
        Pre-allocated, but empty; any values in this field are discarded.
        Changed in-place. Local for each MPI rank--the full computaional grid
        is represented by the collective fields of all MPI ranks.
    compute_potential : bool, optional
        If :code:`True`, a :code:`D+1`-th copy of the Fourier transformed
        external potential field is made to be used later in particle-field
        energy calculation. If :code:`False`, only :code:`D` copies are made.

    See also
    --------
    compute_field_and_kinetic_energy :
        Compute the particle-field energy after the external potential is
        calculated.
    """
    V = np.prod(config.box_size)
    n_mesh_cells = np.prod(np.full(3, config.mesh_size))
    volume_per_cell = V / n_mesh_cells
    for t in range(config.n_types):
        pm.paint(positions[types == t], layout=layouts[t], out=phi[t])
        phi[t] /= volume_per_cell
        phi[t].r2c(out=phi_fourier[t])
        phi_fourier[t].apply(hamiltonian.H, out=Ellipsis)
        phi_fourier[t].c2r(out=phi[t])

    # External potential
    for t in range(config.n_types):
        hamiltonian.v_ext[t](phi).r2c(out=v_ext_fourier[0])
        v_ext_fourier[0].apply(hamiltonian.H, out=Ellipsis)
        np.copyto(
            v_ext_fourier[1].value, v_ext_fourier[0].value, casting="no",
            where=True,
        )
        np.copyto(
            v_ext_fourier[2].value, v_ext_fourier[0].value, casting="no",
            where=True,
        )
        if compute_potential:
            np.copyto(
                v_ext_fourier[3].value, v_ext_fourier[0].value, casting="no",
                where=True,
            )

        # Differentiate the external potential in fourier space
        for d in range(3):
            def force_transfer_function(k, v, d=d):
                return -k[d] * 1j * v

            v_ext_fourier[d].apply(force_transfer_function, out=Ellipsis)
            v_ext_fourier[d].c2r(out=force_mesh[t][d])

        if compute_potential:
            v_ext_fourier[3].c2r(out=v_ext[t])


def compute_field_and_kinetic_energy(
    phi, velocity, hamiltonian, positions, types, v_ext, config, layouts,
    comm=MPI.COMM_WORLD,
):
    """Compute the particle-field and kintic energy contributions

    Calculates the kinetic energy through

    .. math::

        E_k = \\sum_j \\frac{1}{2}m_j\\mathbf{v}_j\\cdot\\mathbf{v}_j,

    and the particle-field energy by

    .. math::

        E_\\text{field} = \\int\\mathrm{d}\\mathbf{r}\\,
            w[\\tilde\\phi],

    where :math:`w` is the interaction energy functional `density`.

    Parameters
    ----------
    phi : list[pmesh.pm.RealField]
        Pmesh :code:`RealField` objects containing discretized particle number
        density values on the computational grid; one for each particle type.
        Local for each MPI rank--the full computaional grid is represented by
        the collective fields of all MPI ranks.
    velocity : (N,D) numpy.ndarray
        Array of velocities for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    hamiltonian : Hamiltonian
        Particle-field interaction energy handler object. Defines the
        grid-independent filtering function, :math:`H`.
    positions : (N,D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    types : (N,) numpy.ndarray
        Array of type indices for each of :code:`N` particles. Local for each
        MPI rank.
    v_ext : list[pmesh.pm.RealField]
        Pmesh :code:`RealField` objects containing discretized particle-field
        external potential values on the computational grid; one for each
        particle type. Local for each MPI rank--the full computaional grid is
        represented by the collective fields of all MPI ranks.
    config : Config
        Configuration object.
    layouts : list[pmesh.domain.Layout]
        Pmesh communication layout objects for domain decompositions of each
        particle type. Used as blueprint by :code:`pmesh.pm.readout` for
        exchange of particle information across MPI ranks as necessary.
    comm : mpi4py.Comm
        MPI communicator to use for rank commuication.

    See also
    --------
    update_field :
        Computes the up-to-date external potential for use in calculating the
        particle-field energy.
    """
    V = np.prod(config.box_size)
    n_mesh__cells = np.prod(np.full(3, config.mesh_size))
    volume_per_cell = V / n_mesh__cells

    w = hamiltonian.w(phi) * volume_per_cell
    field_energy = w.csum()
    kinetic_energy = comm.allreduce(0.5 * config.mass * np.sum(velocity ** 2))
    return field_energy, kinetic_energy


def domain_decomposition(
    positions, pm, *args, molecules=None, bonds=None, verbose=0,
    comm=MPI.COMM_WORLD
):
    """Performs domain decomposition

    Rearranges the MPI rank memory layout into one that better mimics the PFFT
    pencil grid 2D decomposition. As molecules are always required to be fully
    contained on a single MPI rank, a perfect decomposition is not always
    possible. The best decomposition which leaves the center of mass of all
    molecules in their respective correct domains is used, with possible
    overlap into neighbouring domains if the spatial extent of any molecule
    crosses domain boundaries.

    Parameters
    ----------
    positions : (N,D) numpy.ndarray
        Array of positions for :code:`N` particles in :code:`D` dimensions.
        Local for each MPI rank.
    pm : pmesh.pm.ParticleMesh
        Pmesh :code:`ParticleMesh` object interfacing to the CIC window
        function and the PFFT discrete Fourier transform library.
    *args
        Variable length argument list containing arrays to include in the
        domain decomposition.
    molecules : (N,) numpy.ndarray, optional
        Array of integer molecule affiliation for each of :code:`N` particles.
        Global (across all MPI ranks) or local (local indices on this MPI rank
        only) may be used, both, without affecting the result.
    bonds : (N,M) numpy.ndarray, optional
        Array of :code:`M` bonds originating from each of :code:`N` particles.
    verbose : int, optional
        Specify the logging event verbosity of this function.
    comm : mpi4py.Comm
        MPI communicator to use for rank commuication.

    Returns
    -------
    Domain decomposed :code:`positions` and any array specified in
    :code:`*args`, in addition to :code:`bonds` and :code:`molecules` if these
    arrays were provided as input.
    """
    if molecules is not None:
        assert bonds is not None, "bonds must be provided with molecules"
        unique_molecules = np.sort(np.unique(molecules))
        molecules_com = np.empty_like(positions)
        for m in unique_molecules:
            ind = molecules == m
            r = positions[ind, :][0, :]
            molecules_com[ind, :] = r
        layout = pm.decompose(molecules_com, smoothing=0)
        args = (*args, bonds, molecules)
    else:
        layout = pm.decompose(positions, smoothing=0)
    if verbose > 1:
        Logger.rank0.log(
            logging.INFO,
            "DOMAIN_DECOMP: Total number of particles to be exchanged = %d",
            np.sum(layout.get_exchange_cost()),
        )
    return layout.exchange(positions, *args)



"""Parses and handles the configuration information provided for the simulation
"""
import copy
import tomli
import datetime
import logging
import warnings
import numpy as np
from mpi4py import MPI
from dataclasses import dataclass, field
from typing import List, Union, ClassVar
from .force import Bond, Angle, Dihedral, Chi
from .logger import Logger


@dataclass
class Config:
    """Configuration object

    Handles and verifies the simulation configuration specified in the
    configuration file.

    Attributes
    ----------
    gas_constant : float
        Constant value of the gas constant, R (equivalently the Boltzmann
        constant) in the units used internally in HyMD.
    coulomb_constant : float
        Constant value of the Coulomb constant which converts electric field
        values to forces and electric potential values to energies in the units
        used internally in HyMD.
    n_steps: int
        Number of time steps in the simulation.
    time_step: float
        *Outer* time step used in the simulation. If the rRESPA intergrator is
        used, the *inner* time step (the time step used in the integration of
        intramolecular bending, stretching, and torsional forces) is
        :code:`time_step / respa_inner`.
    box_size : list[float] or (D,) numpy.ndarray
        Simulation box size of simulation in :code:`D` dimensions in units of
        nanometers.
    mesh_size : list[int] or int or numpy.ndarray
        Number of grid points used for the discrete density grid.
    sigma : float
        Filter width representing the effective coarse-graining level of the
        particles in the simulation.
    kappa : float
        Compressibility parameter used in the relaxed incompressibility term in
        the interaction energy functional.
    n_print : int, optional
        Frequency of trajectory/energy output to the H5MD trajectory/energy
        output file (in units of number of time steps).
    tau : float, optional
        The time scale of the CSVR thermostat coupling.
    start_temperature : float, optional
        Generate starting temperature by assigning all particle velocities
        randomly according to the Maxwell-Boltzmann distribution at
        :code:`start_temperature` Kelvin prior to starting the simulation.
    target_temperature : float, optional
        Couple the system to a heat bath at :code:`target_temperature` Kelvin.
    mass : float, optional
        Mass of the particles in the simulation.
    hamiltonian : str, optional
        Specifies the interaction energy functional :math:`W[\\tilde\\phi]`
        for use with the particle-field interactions. Options:
        :code:`SquaredPhi`, :code:`DefaultNohChi`, or :code:`DefaultWithChi`.
    domain_decomposition : int, optional
        Specifies the interval (in time steps) of domain decomposition
        exchange, involving all MPI ranks sending and receiving particles
        according to the particles’ positions in the integration box and the
        MPI ranks’ assigned domain.
    integrator : str, optional
        Specifies the time integrator used in the simulation. Options:
        :code:`velocity-verlet` or :code:`respa`.
    respa_inner : int, optional
        The number of inner time steps in the rRESPA integrator. This denotes
        the number of intramolecular force calculations (stretching, bending,
        torsional) are performed between each impulse applied from the field
        forces.
    file_name : str, optional
        File path of the parsed configuration file.
    name : str, optional
        Name for the simulation.
    tags : list[str], optional
        Tags for the simulation.
    bonds : list[Bond], optional
        Specifies harmonic stretching potentials between particles in the
        same molecule.
    angle_bonds : list[Angle], optional
        Specifies harmonic angular bending potentials between particles in the
        same molecule.
    dihedrals : list[Dihedral], optional
        Specifies four-particle torsional potentials by cosine series.
    chi : list[Chi], optional
        Specifies :math:`\\chi`-interaction parameters between particle
        species.
    n_particles : int, optional
        Specifies the total number of particles in the input. Optional keyword
        for validation, ensuring the input HDF5 topology has the correct number
        of particles and molecules.
    max_molecule_size : int, optional
        Maximum size of any single molecule in the system. Used to speed up
        distribution of particles onto MPI ranks in a parallel fashion.
    n_flush : int, optional
        Frequency of HDF5 write buffer flush, forcing trajectory/energy to be
        written to disk (in units of number of :code:`n_print`).
    thermostat_work : float
        Work performed by the thermostat on the system.
    thermostat_coupling_groups : list[str], optional
        Specifies individual groups coupling independently to the CSVR
        thermostat. E.g. in a system containing :code:`"A"`, :code:`"B"`, and
        :code:`"C"` type particles,
        :code:`thermostat_coupling_groups = [["A", "B"], ["C"],]` would
        thermalise types :code:`"A"` and :code:`"B"` together and couple
        :code:`"C"` type particles to a different thermostat (all individual
        thermostats are at the same temperature, i.e. target_temperature
        Kelvin).
    initial_energy : float
        Value of the total energy prior to the start of the simulation.
    cancel_com_momentum : int, optional
        If :code:`True`, the total linear momentum of the center of mass is
        removed before starting the simulation. If an integer is specifed, the
        total linear momentum of the center of mass is removed every
        :code:`remove_com_momentum` time steps. If :code:`False`, the linear
        momentum is never removed.
    coulombtype : str, optional
        Specifies the type of electrostatic Coulomb interactions in the system.
        The strength of the electrostatic forces is modulated by the relative
        dielectric constant of the simulation medium, specified with the
        :code:`dielectric_const` keyword. Charges for individual particles are
        specified in the structure/topology HDF5 input file, not in the
        configuration file. If no charges (or peptide backbone dipoles) are
        present, the electrostatic forces will not be calculated even if this
        keyword is set to :code:`"PIC_Spectral"`.
    dielectric_const : float, optional
        Specifies the relative dielectric constant of the simulation medium
        which regulates the strength of the electrostatic interactions. When
        using helical propensity dihedrals, this keyword must be specified—even
        if electrostatics are not included with the :code:`coulombtype`
        keyword.

    See also
    --------
    hymd.input_parser.Bond :
        Two-particle bond type dataclass.
    hymd.input_parser.Angle :
        Three-particle bond type dataclass.
    hymd.input_parser.Dihedral :
        Four-particle bond type dataclass.
    """
    gas_constant: ClassVar[float] = 0.0083144621  # kJ mol-1 K-1
    coulomb_constant: ClassVar[float] = 138.935458  # kJ mn mol-1 e-2

    n_steps: int
    time_step: float
    box_size: Union[List[float], np.ndarray]
    mesh_size: Union[Union[List[int], np.ndarray], int]
    sigma: float
    kappa: float

    n_print: int = None
    tau: float = None
    start_temperature: Union[float, bool] = None
    target_temperature: Union[float, bool] = None
    mass: float = None
    hamiltonian: str = None
    domain_decomposition: Union[int, bool] = None
    integrator: str = None
    respa_inner: int = 1
    file_name: str = "<config file path unknown>"
    name: str = None
    tags: List[str] = field(default_factory=list)
    bonds: List[Bond] = field(default_factory=list)
    angle_bonds: List[Angle] = field(default_factory=list)
    dihedrals: List[Dihedral] = field(default_factory=list)
    chi: List[Chi] = field(default_factory=list)
    n_particles: int = None
    max_molecule_size: int = None
    n_flush: int = None
    thermostat_work: float = 0.0
    thermostat_coupling_groups: List[List[str]] = field(default_factory=list)
    initial_energy: float = None
    cancel_com_momentum: Union[int, bool] = False
    coulombtype: str = None
    dielectric_const: float = None

    def __str__(self):
        bonds_str = "\tbonds:\n" + "".join(
            [
                (
                    f"\t\t{k.atom_1} {k.atom_2}: "
                    f"{k.equilibrium}, {k.strength}\n"
                )
                for k in self.bonds
            ]
        )
        angle_str = "\tangle_bonds:\n" + "".join(
            [
                (
                    f"\t\t{k.atom_1} {k.atom_2} {k.atom_3}: "
                    + f"{k.equilibrium}, {k.strength}\n"
                )
                for k in self.angle_bonds
            ]
        )
        dihedrals_str = "\tdihedrals:\n" + "".join(
            [
                (
                    f"\t\t{k.atom_1} {k.atom_2} {k.atom_3} {k.atom_4}: "
                    # This might need to be fixed/made prettier, probably
                    # there's an easier way
                    + (
                        "\n\t\t"
                        + " " * len(
                            f"{k.atom_1} {k.atom_2} {k.atom_3} {k.atom_4}: "
                        )
                    ).join(
                        map(
                            str,
                            [
                                [round(num, 3) for num in c_in]
                                if isinstance(c_in, list)
                                else c_in
                                for c_in in k.coeffs
                            ],
                        )
                    )
                    + (
                        "\n\t\t"
                        + " " * len(
                            f"{k.atom_1} {k.atom_2} {k.atom_3} {k.atom_4}: "
                        )
                    )
                    + f"dih_type = {k.dih_type}\n"
                )
                for k in self.dihedrals
            ]
        )
        chi_str = "\tchi:\n" + "".join(
            [
                (f"\t\t{k.atom_1} {k.atom_2}: " + f"{k.interaction_energy}\n")
                for k in self.chi
            ]
        )
        thermostat_coupling_groups_str = ""
        if any(self.thermostat_coupling_groups):
            thermostat_coupling_groups_str = (
                "\tthermostat_coupling_groups:\n"
                + "".join(
                    [
                        "\t\t" + ", ".join([f"{n}" for n in ng]) + "\n"
                        for ng in self.thermostat_coupling_groups
                    ]
                )
            )

        ret_str = f'\n\n\tConfig: {self.file_name}\n\t{50 * "-"}\n'
        for k, v in self.__dict__.items():
            if k not in (
                "bonds",
                "angle_bonds",
                "dihedrals",
                "chi",
                "thermostat_coupling_groups",
            ):
                ret_str += f"\t{k}: {v}\n"
        ret_str += (
            bonds_str
            + angle_str
            + dihedrals_str
            + chi_str
            + thermostat_coupling_groups_str
        )
        return ret_str


def read_config_toml(file_path):
    with open(file_path, "r") as in_file:
        toml_content = in_file.read()
    return toml_content


def propensity_potential_coeffs(x: float, comm):
    alpha_coeffs = np.array(
        [
            [7.406, -5.298, -2.570, 1.336, 0.739],
            [-0.28632126, 1.2099146, 1.18122138, 0.49075168, 0.98495911],
        ]
    )
    beta_coeffs = np.array(
        [
            [3.770, 5.929, -4.151, -0.846, 0.190],
            [-0.2300693, -0.0583289, 0.99342396, 1.03237971, 2.90160988],
        ]
    )
    coil_coeffs = np.array(
        [
            [1.416, -0.739, 0.990, -0.397, 0.136],
            [1.3495933, 0.45649087, 2.30441057, -0.12274901, -0.26179939],
        ]
    )

    zero_add = np.zeros((2, 5))
    if x == -1:
        return np.concatenate((alpha_coeffs, zero_add))
    elif x == 0:
        return np.concatenate((coil_coeffs, zero_add))
    elif x == 1:
        return np.concatenate((beta_coeffs, zero_add))

    abs_x = np.abs(x)
    if abs_x > 1:
        err_str = (
            f"The provided value of λ = {x} is out of λ definition range, "
            f"[-1.0, 1.0]."
        )
        Logger.rank0.log(logging.ERROR, err_str)
        if comm.Get_rank() == 0:
            raise ValueError(err_str)

    else:
        coil_coeffs[0] *= 1 - abs_x
        if x < 0:
            alpha_coeffs[0] *= 0.5 * (abs_x - x)
            return np.concatenate((alpha_coeffs, coil_coeffs))
        else:
            beta_coeffs[0] *= 0.5 * (abs_x + x)
            return np.concatenate((beta_coeffs, coil_coeffs))


def parse_config_toml(toml_content, file_path=None, comm=MPI.COMM_WORLD):
    parsed_toml = tomli.loads(toml_content)
    config_dict = {}

    # Defaults = None
    for n in (
        "n_print",
        "tau",
        "start_temperature",
        "target_temperature",
        "mass",
        "hamiltonian",
        "domain_decomposition",
        "integrator",
        "name",
        "n_particles",
        "max_molecule_size",
        "coulombtype",
        "dielectric_const",
    ):
        config_dict[n] = None

    # Defaults = []
    for n in ("bonds", "angle_bonds", "dihedrals", "chi", "tags"):
        config_dict[n] = []

    # Flatten the .toml dictionary, ignoring the top level [tag] directives (if
    # any).
    for k, v in parsed_toml.items():
        if isinstance(v, dict):
            for nested_k, nested_v in v.items():
                config_dict[nested_k] = nested_v
        else:
            config_dict[k] = v
    for k, v in config_dict.items():
        if k == "bonds":
            config_dict["bonds"] = [None] * len(v)
            for i, b in enumerate(v):
                config_dict["bonds"][i] = Bond(
                    atom_1=b[0],
                    atom_2=b[1],
                    equilibrium=b[2],
                    strength=b[3],
                )
        if k == "angle_bonds":
            config_dict["angle_bonds"] = [None] * len(v)
            for i, b in enumerate(v):
                config_dict["angle_bonds"][i] = Angle(
                    atom_1=b[0],
                    atom_2=b[1],
                    atom_3=b[2],
                    equilibrium=b[3],
                    strength=b[4],
                )
        if k == "dihedrals":
            config_dict["dihedrals"] = [None] * len(v)
            for i, b in enumerate(v):
                try:
                    dih_type = int(b[2][0])
                except IndexError:
                    Logger.rank0.log(
                        logging.WARNING,
                        "Dihedral type not provided, defaulting to 0."
                    )
                    dih_type = 0

                    # Probably it's better to move this in check_dihedrals?
                    wrong_len = len(b[1]) not in (1, 2)
                    wrong_type_1 = len(b[1]) == 1 and not isinstance(
                        b[1][0], float
                    )
                    wrong_type_2 = len(b[1]) == 2 and not isinstance(
                        b[1][0], list
                    )
                    if wrong_len or wrong_type_1 or wrong_type_2:
                        err_str = (
                            "The coefficients specified for the dihedral type "
                            "(0) do not match the correct structure. Either "
                            "use [lambda] or [[cn_prop], [dn_prop]], or select"
                            " the correct dihedral type."
                        )
                        Logger.rank0.log(logging.ERROR, err_str)
                        if comm.Get_rank() == 0:
                            raise RuntimeError(err_str)

                # FIXME: this is messy af, I don't like it
                if dih_type == 0 and isinstance(b[1][0], (float, int)):
                    coeff = propensity_potential_coeffs(b[1][0], comm)
                elif dih_type == 1 and len(b[1]) == 3:
                    coeff = np.array(
                        propensity_potential_coeffs(b[1][0][0], comm).tolist()
                        + b[1][1:]
                    )
                elif dih_type == 2:
                    coeff = np.array(b[1])
                else:
                    coeff = np.insert(
                        np.array(b[1]), 2, np.zeros((2, 5)), axis=0
                    )

                config_dict["dihedrals"][i] = Dihedral(
                    atom_1=b[0][0],
                    atom_2=b[0][1],
                    atom_3=b[0][2],
                    atom_4=b[0][3],
                    coeffs=coeff,
                    dih_type=dih_type,
                )
        # if k == "improper dihedrals":
        #     config_dict["improper dihedrals"] = [None] * len(v)
        # ...
        if k == "chi":
            config_dict["chi"] = [None] * len(v)
            for i, c in enumerate(v):
                c_ = sorted([c[0], c[1]])
                config_dict["chi"][i] = Chi(
                    atom_1=c_[0], atom_2=c_[1], interaction_energy=c[2]
                )

    if file_path is not None:
        config_dict["file_name"] = file_path

    for n in (
        "n_steps", "time_step", "box_size", "mesh_size", "sigma", "kappa"
    ):
        if n not in config_dict:
            err_str = (
                f"No {n} specified in config file {file_path}. Unable to start"
                f" simulation."
            )
            Logger.rank0.log(logging.ERROR, err_str)
            if comm.Get_rank() == 0:
                raise ValueError(err_str)
    return Config(**config_dict)


def check_n_particles(config, indices, comm=MPI.COMM_WORLD):
    n_particles = comm.allreduce(len(indices), MPI.SUM)
    if config.n_particles is None:
        config = copy.deepcopy(config)
        config.n_particles = n_particles
        info_str = (
            f"No n_particles found in toml file {config.file_name}, defaulting"
            f" to indices.shape ({n_particles})"
        )
        Logger.rank0.log(logging.INFO, info_str)
        return config

    if n_particles != config.n_particles:
        warn_str = (
            f"n_particles in {config.file_name} ({config.n_particles}) does "
            "not match the shape of the indices array in the .HDF5 file "
            f"({n_particles}). Defaulting to using indices.shape "
            f"({n_particles})"
        )
        Logger.rank0.log(logging.WARNING, warn_str)
        if comm.Get_rank() == 0:
            warnings.warn(warn_str)
        config = copy.deepcopy(config)
        config.n_particles = n_particles
    return config


def check_max_molecule_size(config, comm=MPI.COMM_WORLD):
    if config.max_molecule_size is None:
        info_str = (
            f"No max_molecule_size found in toml file {config.file_name}, "
            f"defaulting to 201"
        )
        Logger.rank0.log(logging.INFO, info_str)
        config = copy.deepcopy(config)
        config.max_molecule_size = 201
        return config

    if config.max_molecule_size < 1:
        warn_str = (
            f"max_molecule_size in {config.file_name} must be a positive "
            f"integer, not {config.max_molecule_size}, defaulting to 201"
        )
        Logger.rank0.log(logging.WARNING, warn_str)
        if comm.Get_rank() == 0:
            warnings.warn(warn_str)
        config = copy.deepcopy(config)
        config.max_molecule_size = 201
        return config
    return config


def _find_unique_names(config, names, comm=MPI.COMM_WORLD):
    unique_names = np.unique(names)
    receive_buffer = comm.gather(unique_names, root=0)

    gathered_unique_names = None
    if comm.Get_rank() == 0:
        gathered_unique_names = np.unique(np.concatenate(receive_buffer))
    unique_names = comm.bcast(gathered_unique_names, root=0)
    unique_names = sorted([n.decode("UTF-8") for n in unique_names])
    config.unique_names = unique_names
    config.n_types = len(unique_names)
    return config


def _setup_type_to_name_map(config, names, types, comm=MPI.COMM_WORLD):
    if not hasattr(config, "unique_names"):
        config = _find_unique_names(config, names)
    name_to_type_ = {}
    for n, t in zip(names, types):
        n = n.decode("utf-8")
        if n not in name_to_type_:
            name_to_type_[n] = t
    receive_buffer = comm.gather(name_to_type_, root=0)
    gathered_dict = None
    if comm.Get_rank() == 0:
        gathered_dict = {}
        for d in receive_buffer:
            for k, v in d.items():
                if k not in gathered_dict:
                    gathered_dict[k] = v
                else:
                    assert v == gathered_dict[k]
    name_to_type_map = comm.bcast(gathered_dict, root=0)
    config.name_to_type_map = name_to_type_map
    config.type_to_name_map = {v: k for k, v in name_to_type_map.items()}
    return config


def check_bonds(config, names, comm=MPI.COMM_WORLD):
    if not hasattr(config, "unique_names"):
        config = _find_unique_names(config, names)
    unique_names = config.unique_names

    for b in config.bonds:
        if b.atom_1 not in unique_names or b.atom_2 not in unique_names:
            missing_str = ""
            if b.atom_1 not in unique_names:
                if b.atom_2 not in unique_names:
                    if b.atom_1 == b.atom_2:
                        missing_str = f"no {b.atom_1} atoms"
                    else:
                        missing_str = f"neither {b.atom_1}, {b.atom_2} atoms"
                else:
                    missing_str = f"no {b.atom_1} atoms"
            else:
                missing_str = f"no {b.atom_2} atoms"

            warn_str = (
                f"Bond type {b.atom_1}--{b.atom_2} specified in "
                f"{config.file_name} but {missing_str} are present in the "
                f"specified system (names array)"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
    return config


def check_angles(config, names, comm=MPI.COMM_WORLD):
    if not hasattr(config, "unique_names"):
        config = _find_unique_names(config, names)
    unique_names = config.unique_names

    for a in config.angle_bonds:
        if (
            a.atom_1 not in unique_names
            or a.atom_2 not in unique_names
            or a.atom_3 not in unique_names
        ):
            missing = [
                a.atom_1 not in unique_names,
                a.atom_2 not in unique_names,
                a.atom_3 not in unique_names,
            ]
            missing_names = [
                atom
                for i, atom in enumerate([a.atom_1, a.atom_2, a.atom_3])
                if missing[i]
            ]
            missing_str = ", ".join(np.unique(missing_names))

            warn_str = (
                f"Angle bond type {a.atom_1}--{a.atom_2}--{a.atom_3} "
                f"specified in {config.file_name} but no {missing_str} atoms "
                f"are present in the specified system (names array)"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
    return config


def check_dihedrals(config, names, comm=MPI.COMM_WORLD):
    if not hasattr(config, "unique_names"):
        config = _find_unique_names(config, names)
    unique_names = config.unique_names

    for d in config.dihedrals:
        if (
            d.atom_1 not in unique_names
            or d.atom_2 not in unique_names
            or d.atom_3 not in unique_names
            or d.atom_4 not in unique_names
        ):
            missing = [
                d.atom_1 not in unique_names,
                d.atom_2 not in unique_names,
                d.atom_3 not in unique_names,
                d.atom_4 not in unique_names,
            ]
            missing_names = [
                atom
                for i, atom in enumerate(
                    [d.atom_1, d.atom_2, d.atom_3, d.atom_4]
                )
                if missing[i]
            ]
            missing_str = ", ".join(np.unique(missing_names))

            warn_str = (
                f"Dihedral type {d.atom_1}--{d.atom_2}--{d.atom_3}--{d.atom_4}"
                f" specified in {config.file_name} but no {missing_str} atoms "
                f"are present in the specified system (names array)"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
    return config


def check_chi(config, names, comm=MPI.COMM_WORLD):
    if not hasattr(config, "unique_names"):
        config = _find_unique_names(config, names)
    unique_names = config.unique_names

    for c in config.chi:
        if c.atom_1 not in unique_names or c.atom_2 not in unique_names:
            missing_str = ""
            if c.atom_1 not in unique_names:
                if c.atom_2 not in unique_names:
                    if c.atom_1 == c.atom_2:
                        missing_str = f"no {c.atom_1} atoms"
                    else:
                        missing_str = f"neither {c.atom_1}, {c.atom_2} atoms"
                else:
                    missing_str = f"no {c.atom_1} atoms"
            else:
                missing_str = f"no {c.atom_2} atoms"

            warn_str = (
                f"Chi interaction {c.atom_1}--{c.atom_2} specified in "
                f"{config.file_name} but {missing_str} are present in the "
                f"specified system (names array)"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)

    for i, n in enumerate(unique_names):
        for m in unique_names[i+1:]:
            found = False
            for c in config.chi:
                if (c.atom_1 == n and c.atom_2 == m) or (
                    c.atom_1 == m and c.atom_2 == n
                ):
                    found = True
            if not found:
                config.chi.append(
                    Chi(atom_1=n, atom_2=m, interaction_energy=0.0)
                )
                warn_str = (
                    f"Atom types {n} and {m} found in the "
                    f"system, but no chi interaction {n}--{m} "
                    f"specified in {config.file_name}. Defaulting to "
                    f"chi[{n}, {m}] = 0"
                )
                Logger.rank0.log(logging.WARNING, warn_str)
                if comm.Get_rank() == 0:
                    warnings.warn(warn_str)
    return config


def check_box_size(config, comm=MPI.COMM_WORLD):
    for b in config.box_size:
        if b <= 0.0:
            err_str = (
                f"Invalid box size specified in {config.file_name}: "
                f"{config.box_size}"
            )
            Logger.rank0.log(logging.ERROR, err_str)
            if comm.Get_rank() == 0:
                raise ValueError(err_str)
    config.box_size = np.array(config.box_size, dtype=np.float64)
    return config


def check_integrator(config, comm=MPI.COMM_WORLD):
    if config.integrator.lower() not in ("velocity-verlet", "respa"):
        err_str = (
            f"Invalid integrator specified in {config.file_name}: "
            f'{config.integrator}. Available options "velocity-verlet" or '
            f'"respa".'
        )
        Logger.rank0.log(logging.ERROR, err_str)
        if comm.Get_rank() == 0:
            raise ValueError(err_str)

    if config.integrator.lower() == "respa":
        if not isinstance(config.respa_inner, int):
            if isinstance(config.respa_inner, float):
                if config.respa_inner.is_int():
                    warn_str = (
                        f"Number of inner rRESPA time steps in "
                        f"{config.file_name}: {config.respa_inner} specified "
                        f"as float, using {int(config.respa_inner)}"
                    )
                    Logger.rank0.log(logging.WARNING, err_str)
                    if comm.Get_rank() == 0:
                        warnings.warn(warn_str)
                    config.respa_inner = int(config.respa_inner)
                else:
                    err_str = (
                        f"Invalid number of inner rRESPA time steps in "
                        f"{config.file_name}: {config.respa_inner}. Must be "
                        f"positive integer"
                    )
                    Logger.rank0.log(logging.ERROR, err_str)
                    if comm.Get_rank() == 0:
                        raise ValueError(err_str)
            else:
                err_str = (
                    f"Invalid number of inner rRESPA time steps in "
                    f"{config.file_name}: {config.respa_inner}. Must be "
                    f"positive integer"
                )
                Logger.rank0.log(logging.ERROR, err_str)
                if comm.Get_rank() == 0:
                    raise TypeError(err_str)

    if (
        config.integrator.lower() == "velocity-verlet"
        and config.respa_inner != 1
    ):
        warn_str = (
            f"Integrator type Velocity-Verlet specified in {config.file_name} "
            f"and inner rRESPA time steps set to {config.respa_inner}. "
            f"Using respa_inner = 1"
        )
        Logger.rank0.log(logging.WARNING, warn_str)
        if comm.Get_rank() == 0:
            warnings.warn(warn_str)
        config.respa_inner = 1

    return config


def check_hamiltonian(config, comm=MPI.COMM_WORLD):
    if config.hamiltonian is None:
        if len(config.chi) == 0:
            warn_str = (
                f"No hamiltonian form and no chi interactions specified in "
                f"{config.file_name}, defaulting to DefaultNoChi hamiltonian"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
            config.hamiltonian = "DefaultNoChi"
        else:
            warn_str = (
                f"No hamiltonian form specified in {config.file_name}, but "
                f"chi interactions are specified, defaulting to "
                f"DefaultWithChi hamiltonian"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
            config.hamiltonian = "DefaultWithChi"
    return config


def check_n_print(config, comm=MPI.COMM_WORLD):
    if config.n_print is None or config.n_print < 0:
        config.n_print = False
        return config
    elif not isinstance(config.n_print, int):
        if isinstance(config.n_print, float) and config.n_print.is_int():
            config.n_print = int(config.n_print)
        elif isinstance(config.n_print, float):
            warn_str = (
                f"n_print is a float ({config.n_print}), not int, using "
                f"{int(round(config.n_print))}"
            )
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)

            n_print = int(round(config.n_print))
            if n_print > 0:
                config.n_print = n_print
            else:
                config.n_print = False
    return config


def check_tau(config, comm=MPI.COMM_WORLD):
    if config.tau is None and config.target_temperature is not None:
        warn_str = "target temp specified but no tau, defaulting 0.7"
        config.tau = 0.7
        Logger.rank0.log(logging.WARNING, warn_str)
        if comm.Get_rank() == 0:
            warnings.warn(warn_str)
    return config


def check_start_and_target_temperature(config, comm=MPI.COMM_WORLD):
    """Validate provided starting and target thermostat temperatures

    Assesses the provided temperature target and ensures it is a non-negative
    floating point number or :code:`False`. Ensures the starting temperature is
    a non-negative floating point number or :code:`False`.

    If the value for either is :code:`None`, the returned configuration object
    has the values defaulted to :code:`False` in each case.

    Parameters
    ----------
    config : Config
        Configuration object.

    Returns
    -------
    validated_config : Config
        Configuration object with validated :code:`target_temperature` and
        :code:`start_temperature`.
    """
    for t in ("start_temperature", "target_temperature"):
        if getattr(config, t) is not None:
            try:
                if getattr(config, t) < 0:
                    warn_str = (
                        f"{t} set to negative value ({getattr(config, t)}), "
                        f"defaulting to False"
                    )
                    setattr(config, t, False)
                    Logger.rank0.log(logging.WARNING, warn_str)
                    if comm.Get_rank() == 0:
                        warnings.warn(warn_str)
            except TypeError as e:
                err_str = (
                    f"Could not interpret {t} = {repr(getattr(config, t))} as "
                    f"a number."
                )
                raise TypeError(err_str) from e

    if config.start_temperature is None:
        config.start_temperature = False
    if config.target_temperature is None:
        config.target_temperature = False
    return config


def check_mass(config, comm=MPI.COMM_WORLD):
    if config.mass is not None:
        try:
            config.mass = float(config.mass)
            return config
        except ValueError:
            pass

    if config.mass is None:
        info_str = "no mass specified, defaulting to 72.0"
        config.mass = 72.0
        Logger.rank0.log(logging.INFO, info_str)
    elif isinstance(config.mass, int) or isinstance(config.mass, float):
        err_str = (
            f"specified mass is invalid type {config.mass}, "
            f"({type(config.mass)})"
        )
        Logger.rank0.log(logging.ERROR, err_str)
        if comm.Get_rank() == 0:
            raise TypeError(err_str)
    elif config.mass < 0:
        err_str = f"invalid mass specified, {config.mass}"
        Logger.rank0.log(logging.ERROR, err_str)
        if comm.Get_rank() == 0:
            raise TypeError(err_str)
    return config


def check_domain_decomposition(config, comm=MPI.COMM_WORLD):
    if config.domain_decomposition is None:
        config.domain_decomposition = False
    dd = config.domain_decomposition
    assert isinstance(dd, int) or isinstance(dd, float) or (dd is None)
    if isinstance(dd, int):
        if dd < 0:
            warn_str = "negative domain_decomposition specified, using False"
            config.domain_decomposition = False
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
    if isinstance(dd, float):
        if not dd.is_int():
            warn_str = (
                f"domain_decomposition ({config.domain_decomposition})is not "
                f"an integer, using {int(round(dd))}"
            )
            config.domain_decomposition = int(round(dd))
            Logger.rank0.log(logging.WARNING, warn_str)
            if comm.Get_rank() == 0:
                warnings.warn(warn_str)
    return config


def check_name(config, comm=MPI.COMM_WORLD):
    if config.name is None:
        root_current_time = ""
        if comm.Get_rank() == 0:
            root_current_time = datetime.datetime.now().strftime(
                "%m/%d/%Y, %H:%M:%S"
            )
        current_time = comm.bcast(root_current_time, root=0)

        if config.name is None:
            config.name = "sim" + current_time
    return config


def check_thermostat_coupling_groups(config, comm=MPI.COMM_WORLD):
    if any(config.thermostat_coupling_groups):
        found = [0 for _ in config.unique_names]
        unique_names = [n for n in config.unique_names]
        for i, group in enumerate(config.thermostat_coupling_groups):
            for species in group:
                try:
                    ind = unique_names.index(species)
                    found[ind] += 1
                except ValueError as e:
                    err_str = (
                        f"Particle species {species} specified in thermostat "
                        f"coupling group {i}, but no {species} particles were "
                        "found in the system."
                    )
                    raise ValueError(err_str) from e
        if any([True if f > 1 else False for f in found]):
            for i, f in enumerate(found):
                if f > 1:
                    species = unique_names[i]
                    err_str = (
                        f"Particle species {species} specified in multiple "
                        "thermostat coupling groups; "
                        f"{config.thermostat_coupling_groups}."
                    )
                    raise ValueError(err_str)
        if not all([True if f == 1 else False for f in found]):
            for i, f in enumerate(found):
                if f != 1:
                    species = unique_names[i]
                    err_str = (
                        f"Particle species {species} not specified in any "
                        f"thermostat coupling group, but {species} particles "
                        "were found in the system"
                    )
                    raise ValueError(err_str)
    return config


def check_cancel_com_momentum(config, comm=MPI.COMM_WORLD):
    if isinstance(config.cancel_com_momentum, int):
        if config.cancel_com_momentum == 0 or config.cancel_com_momentum < 0:
            config.cancel_com_momentum = False
    elif not isinstance(config.cancel_com_momentum, bool):
        err_str = (
            f"Could not interpret {config.cancel_com_momentum} as an integer "
            f"or boolean."
        )
        raise ValueError(err_str)
    return config


def check_config(config, indices, names, types, comm=MPI.COMM_WORLD):
    """Performs various checks on the specfied config to ensure consistency

    Parameters
    ----------
    config : Config
        Configuration object.
    indices : (N,) numpy.ndarray
        Array of integer indices for :code:`N` particles.
    names : (N,) numpy.ndarray
        Array of string names for :code:`N` particles.
    types : (N,) numpy.ndarray
        Array of integer type indices for :code:`N` particles.
    comm : mpi4py.Comm, optional
        MPI communicator, defaults to :code:`mpi4py.COMM_WORLD`.

    Returns
    -------
    config : Config
        Validated configuration object.
    """
    config.box_size = np.array(config.box_size)
    config = _find_unique_names(config, names, comm=comm)
    if types is not None:
        config = _setup_type_to_name_map(config, names, types, comm=comm)
    config = check_box_size(config, comm=comm)
    config = check_integrator(config, comm=comm)
    config = check_max_molecule_size(config, comm=comm)
    config = check_tau(config, comm=comm)
    config = check_start_and_target_temperature(config, comm=comm)
    config = check_mass(config, comm=comm)
    config = check_domain_decomposition(config, comm=comm)
    config = check_name(config, comm=comm)
    config = check_n_particles(config, indices, comm=comm)
    config = check_chi(config, names, comm=comm)
    config = check_bonds(config, names, comm=comm)
    config = check_angles(config, names, comm=comm)
    config = check_dihedrals(config, names, comm=comm)
    config = check_hamiltonian(config, comm=comm)
    config = check_thermostat_coupling_groups(config, comm=comm)
    config = check_cancel_com_momentum(config, comm=comm)
    return config



"""Main simulation driver module
"""
import datetime
import h5py
import logging
from mpi4py import MPI
import numpy as np
import pmesh.pm as pmesh
import warnings
from .configure_runtime import configure_runtime
from .hamiltonian import DefaultNoChi, DefaultWithChi, SquaredPhi
from .input_parser import check_config
from .logger import Logger, format_timedelta
from .file_io import distribute_input, OutDataset, store_static, store_data
from .field import (compute_field_force, update_field,
                    compute_field_and_kinetic_energy, domain_decomposition,
                    update_field_force_q, compute_field_energy_q,)
from .thermostat import (csvr_thermostat, cancel_com_momentum,
                         generate_initial_velocities)
from .force import dipole_forces_redistribution, prepare_bonds
from .integrator import integrate_velocity, integrate_position


def main():
    """Main simulation driver

    Initializes structure, topology, and simulation configuration and iterates
    the molecular dynamics loop.
    """
    comm = MPI.COMM_WORLD
    rank = comm.Get_rank()
    size = comm.Get_size()

    if rank == 0:
        start_time = datetime.datetime.now()

    args, config = configure_runtime(comm)

    if args.double_precision:
        dtype = np.float64
        from .force import (
            compute_bond_forces__fortran__double as compute_bond_forces
        )
        from .force import (
            compute_angle_forces__fortran__double as compute_angle_forces
        )
        from .force import (
            compute_dihedral_forces__fortran__double as compute_dihedral_forces
        )
    else:
        dtype = np.float32
        from .force import (
            compute_bond_forces__fortran as compute_bond_forces
        )
        from .force import (
            compute_angle_forces__fortran as compute_angle_forces
        )
        from .force import (
            compute_dihedral_forces__fortran as compute_dihedral_forces
        )

    driver = "mpio" if not args.disable_mpio else None
    _kwargs = {"driver": driver, "comm": comm} if not args.disable_mpio else {}
    with h5py.File(args.input, "r", **_kwargs) as in_file:
        rank_range, molecules_flag = distribute_input(
            in_file,
            rank,
            size,
            config.n_particles,
            config.max_molecule_size if config.max_molecule_size else 201,
            comm=comm,
        )
        indices = in_file["indices"][rank_range]
        positions = in_file["coordinates"][-1, rank_range, :]
        positions = positions.astype(dtype)
        if "velocities" in in_file:
            velocities = in_file["velocities"][-1, rank_range, :]
            velocities = velocities.astype(dtype)
        else:
            velocities = np.zeros_like(positions, dtype=dtype)

        names = in_file["names"][rank_range]

        types = None
        bonds = None
        molecules = []
        if "types" in in_file:
            types = in_file["types"][rank_range]
        if molecules_flag:
            molecules = in_file["molecules"][rank_range]
            bonds = in_file["bonds"][rank_range]
        if "charge" in in_file:
            charges = in_file["charge"][rank_range]
            charges_flag = True
        else:
            charges_flag = False

    config = check_config(config, indices, names, types, comm=comm)
    if config.n_print:
        if config.n_flush is None:
            config.n_flush = 10000 // config.n_print

    if config.start_temperature:
        velocities = generate_initial_velocities(velocities, config, comm=comm)
    elif config.cancel_com_momentum:
        velocities = cancel_com_momentum(velocities, config, comm=comm)

    bond_forces = np.zeros_like(positions)
    angle_forces = np.zeros_like(positions)
    dihedral_forces = np.zeros_like(positions)
    reconstructed_forces = np.zeros_like(positions)
    field_forces = np.zeros_like(positions)
    elec_forces = np.zeros_like(positions)

    positions = np.mod(positions, config.box_size[None, :])

    # Initialize dipoles, populate them if protein_flag == True
    if args.disable_dipole:
        dipole_flag = 0
    else:
        dipole_flag = 1
    protein_flag = 0
    dipole_positions = np.zeros(shape=(4, 3), dtype=dtype)
    dipole_forces = np.zeros(shape=(4, 3))
    dipole_charges = np.zeros(shape=4)
    transfer_matrices = np.zeros(shape=(6, 3, 3), dtype=dtype)

    # Initialize energies
    field_energy = 0.0
    bond_energy = 0.0
    angle_energy = 0.0
    dihedral_energy = 0.0
    kinetic_energy = 0.0
    field_q_energy = 0.0

    # Ignore numpy numpy.VisibleDeprecationWarning: Creating an ndarray from
    # ragged nested sequences until it is fixed in pmesh
    with warnings.catch_warnings():
        warnings.filterwarnings(
            action="ignore",
            category=np.VisibleDeprecationWarning,
            message=r"Creating an ndarray from ragged nested sequences",
        )
        pm = pmesh.ParticleMesh(
            config.mesh_size, BoxSize=config.box_size,
            dtype="f4" if dtype == np.float32 else np.float64, comm=comm
        )

    if config.hamiltonian.lower() == "defaultnochi":
        hamiltonian = DefaultNoChi(config)
    elif config.hamiltonian.lower() == "defaultwithchi":
        hamiltonian = DefaultWithChi(
            config, config.unique_names, config.type_to_name_map
        )
    elif config.hamiltonian.lower() == "squaredphi":
        hamiltonian = SquaredPhi(config)
    else:
        err_str = (
            f"The specified Hamiltonian {config.hamiltonian} was not "
            f"recognized as a valid Hamiltonian."
        )
        Logger.rank0.log(logging.ERROR, err_str)
        if rank == 0:
            raise NotImplementedError(err_str)

    Logger.rank0.log(logging.INFO, f"pfft-python processor mesh: {str(pm.np)}")

    # Initialize density fields
    phi = [pm.create("real", value=0.0) for _ in range(config.n_types)]
    phi_fourier = [
        pm.create("complex", value=0.0) for _ in range(config.n_types)
    ]  # noqa: E501
    force_on_grid = [
        [pm.create("real", value=0.0) for _ in range(3)] for _ in range(config.n_types)  # noqa: E501
    ]
    v_ext_fourier = [pm.create("complex", value=0.0) for _ in range(4)]
    v_ext = [pm.create("real", value=0.0) for _ in range(config.n_types)]

    # Initialize charge density fields
    _SPACE_DIM = 3
    if charges_flag and config.coulombtype == "PIC_Spectral":
        phi_q = pm.create("real", value=0.0)
        phi_q_fourier = pm.create("complex", value=0.0)
        elec_field_fourier = [
            pm.create("complex", value=0.0) for _ in range(_SPACE_DIM)
        ]  # for force calculation
        elec_field = [
            pm.create("real", value=0.0) for _ in range(_SPACE_DIM)
        ]  # for force calculation
        elec_energy_field = pm.create(
            "complex", value=0.0
        )

    args_in = [
        velocities,
        indices,
        bond_forces,
        angle_forces,
        dihedral_forces,
        reconstructed_forces,
        field_forces,
        names,
        types,
    ]
    args_recv = [
        "positions",
        "velocities",
        "indices",
        "bond_forces",
        "angle_forces",
        "dihedral_forces",
        "reconstructed_forces",
        "field_forces",
        "names",
        "types",
    ]

    if charges_flag:
        args_in.append(charges)
        args_in.append(elec_forces)
        args_recv.append("charges")
        args_recv.append("elec_forces")
    if molecules_flag:
        args_recv.append("bonds")
        args_recv.append("molecules")

    _str_receive_dd = ",".join(args_recv)
    _cmd_receive_dd = f"({_str_receive_dd}) = dd"

    if config.domain_decomposition:
        dd = domain_decomposition(
            positions,
            pm,
            *tuple(args_in),
            molecules=molecules if molecules_flag else None,
            bonds=bonds if molecules_flag else None,
            verbose=args.verbose,
            comm=comm,
        )

        exec(_cmd_receive_dd)

    if not args.disable_field:
        layouts = [pm.decompose(positions[types == t]) for t in range(config.n_types)]  # noqa: E501
        update_field(
            phi, layouts, force_on_grid, hamiltonian, pm, positions, types,
            config, v_ext, phi_fourier, v_ext_fourier, compute_potential=True,
        )
        field_energy, kinetic_energy = compute_field_and_kinetic_energy(
            phi, velocities, hamiltonian, positions, types, v_ext, config,
            layouts, comm=comm,
        )
        compute_field_force(
            layouts, positions, force_on_grid, field_forces, types,
            config.n_types
        )
    else:
        kinetic_energy = comm.allreduce(
            0.5 * config.mass * np.sum(velocities**2)
        )

    if charges_flag and config.coulombtype == "PIC_Spectral":
        layout_q = pm.decompose(positions)
        update_field_force_q(
            charges, phi_q, phi_q_fourier, elec_field_fourier, elec_field,
            elec_forces, layout_q, pm, positions, config,
        )

        field_q_energy = compute_field_energy_q(
            config, phi_q_fourier, elec_energy_field, field_q_energy,
            comm=comm,
        )

    if molecules_flag:
        if not (args.disable_bonds
                and args.disable_angle_bonds
                and args.disable_dihedrals):

            bonds_prep = prepare_bonds(
                molecules, names, bonds, indices, config
            )
            (
                # two-particle bonds
                bonds_2_atom1, bonds_2_atom2, bonds_2_equilibrium,
                bonds_2_stength,
                # three-particle bonds
                bonds_3_atom1, bonds_3_atom2, bonds_3_atom3,
                bonds_3_equilibrium, bonds_3_stength,
                # four-particle bonds
                bonds_4_atom1, bonds_4_atom2, bonds_4_atom3, bonds_4_atom4,
                bonds_4_coeff, bonds_4_type, bonds_4_last,
            ) = bonds_prep

            bonds_4_coeff = np.asfortranarray(bonds_4_coeff)

            if bonds_4_type.any() > 1:
                err_str = (
                    "0 and 1 are the only currently supported dihedral angle "
                    "types."
                )
                Logger.rank0.log(logging.ERROR, err_str)
                if rank == 0:
                    raise NotImplementedError(err_str)

            # Check if we have a protein
            protein_flag = comm.allreduce(bonds_4_type.any() == 1)
            if protein_flag and not args.disable_dipole:
                # Each rank will have different n_tors, we do not need to
                # domain decompose dipoles
                n_tors = len(bonds_4_atom1)
                dipole_positions = np.zeros((n_tors, 4, 3), dtype=dtype)
                # 4 because we need to take into account the last angle in the
                # molecule
                dipole_charges = np.array(
                    [
                        2 * [0.25, -0.25]
                        if (bonds_4_type[i], bonds_4_last[i]) == (1, 1)
                        else [0.25, -0.25, 0.0, 0.0]
                        if bonds_4_type[i] == 1
                        else 2 * [0.0, 0.0]
                        for i in range(n_tors)
                    ],
                    dtype=dtype,
                ).flatten()
                dipole_forces = np.zeros_like(dipole_positions)
                transfer_matrices = np.zeros(
                    shape=(n_tors, 6, 3, 3), dtype=dtype
                )
                phi_dipoles = pm.create("real", value=0.0)
                phi_dipoles_fourier = pm.create("complex", value=0.0)
                dipoles_field_fourier = [
                    pm.create("complex", value=0.0) for _ in range(_SPACE_DIM)
                ]
                dipoles_field = [
                    pm.create("real", value=0.0) for _ in range(_SPACE_DIM)
                ]

        positions = np.asfortranarray(positions)
        velocities = np.asfortranarray(velocities)
        bond_forces = np.asfortranarray(bond_forces)
        angle_forces = np.asfortranarray(angle_forces)
        dihedral_forces = np.asfortranarray(dihedral_forces)
        dipole_positions = np.asfortranarray(dipole_positions)
        transfer_matrices = np.asfortranarray(transfer_matrices)

        if not args.disable_bonds:
            bond_energy_ = compute_bond_forces(
                bond_forces, positions, config.box_size, bonds_2_atom1,
                bonds_2_atom2, bonds_2_equilibrium, bonds_2_stength,
            )
            bond_energy = comm.allreduce(bond_energy_, MPI.SUM)
        else:
            bonds_2_atom1, bonds_2_atom2 = [], []
        if not args.disable_angle_bonds:
            angle_energy_ = compute_angle_forces(
                angle_forces, positions, config.box_size, bonds_3_atom1,
                bonds_3_atom2, bonds_3_atom3, bonds_3_equilibrium,
                bonds_3_stength,
            )
            angle_energy = comm.allreduce(angle_energy_, MPI.SUM)
        if not args.disable_dihedrals:
            dihedral_energy_ = compute_dihedral_forces(
                dihedral_forces, positions, dipole_positions,
                transfer_matrices, config.box_size, bonds_4_atom1,
                bonds_4_atom2, bonds_4_atom3, bonds_4_atom4, bonds_4_coeff,
                bonds_4_type, bonds_4_last, dipole_flag,
            )
            dihedral_energy = comm.allreduce(dihedral_energy_, MPI.SUM)

        if protein_flag and not args.disable_dipole:
            dipole_positions = np.reshape(dipole_positions, (4 * n_tors, 3))
            dipole_forces = np.reshape(dipole_forces, (4 * n_tors, 3))

            layout_dipoles = pm.decompose(dipole_positions)
            update_field_force_q(
                dipole_charges, phi_dipoles, phi_dipoles_fourier,
                dipoles_field_fourier, dipoles_field, dipole_forces,
                layout_dipoles, pm, dipole_positions, config,
            )

            dipole_positions = np.reshape(dipole_positions, (n_tors, 4, 3))
            dipole_forces = np.reshape(dipole_forces, (n_tors, 4, 3))

            dipole_positions = np.asfortranarray(dipole_positions)
            dipole_forces_redistribution(
                reconstructed_forces, dipole_forces, transfer_matrices,
                bonds_4_atom1, bonds_4_atom2, bonds_4_atom3, bonds_4_atom4,
                bonds_4_type, bonds_4_last,
            )

    else:
        bonds_2_atom1, bonds_2_atom2 = [], []

    config.initial_energy = (
        field_energy + kinetic_energy + bond_energy + angle_energy
        + dihedral_energy + field_q_energy
    )

    out_dataset = OutDataset(
        args.destdir, config, double_out=args.double_output,
        disable_mpio=args.disable_mpio,
    )

    store_static(
        out_dataset, rank_range, names, types, indices, config, bonds_2_atom1,
        bonds_2_atom2, molecules=molecules if molecules_flag else None,
        velocity_out=args.velocity_output, force_out=args.force_output,
        charges=charges if charges_flag else False, comm=comm,
    )

    if config.n_print > 0:
        step = 0
        frame = 0
        if not args.disable_field:
            field_energy, kinetic_energy = compute_field_and_kinetic_energy(
                phi, velocities, hamiltonian, positions, types, v_ext, config,
                layouts, comm=comm,
            )
        else:
            kinetic_energy = comm.allreduce(
                0.5 * config.mass * np.sum(velocities ** 2)
            )

        temperature = (
            (2 / 3) * kinetic_energy / (config.gas_constant * config.n_particles)  # noqa: E501
        )
        forces_out = (
            field_forces + bond_forces + angle_forces + dihedral_forces
            + reconstructed_forces
        )
        store_data(
            out_dataset, step, frame, indices, positions, velocities,
            forces_out, config.box_size, temperature, kinetic_energy,
            bond_energy, angle_energy, dihedral_energy, field_energy,
            field_q_energy, config.time_step, config,
            velocity_out=args.velocity_output, force_out=args.force_output,
            charge_out=charges_flag, dump_per_particle=args.dump_per_particle,
            comm=comm,
        )

    if rank == 0:
        loop_start_time = datetime.datetime.now()
        last_step_time = datetime.datetime.now()

    # MD loop
    for step in range(config.n_steps):
        current_step_time = datetime.datetime.now()

        if step == 0 and args.verbose > 1:
            Logger.rank0.log(logging.INFO, f"MD step = {step:10d}")
        else:
            log_step = False
            if config.n_steps < 1000:
                log_step = True
            elif (
                np.mod(step, config.n_steps // 1000) == 0
                or np.mod(step, config.n_print) == 0
            ):
                log_step = True
            if rank == 0 and log_step and args.verbose > 1:
                step_t = current_step_time - last_step_time
                tot_t = current_step_time - loop_start_time
                ns_sim = (
                    (step + 1) * config.respa_inner * config.time_step / 1000.0
                )

                seconds_per_day = 24 * 60 * 60
                seconds_elapsed = tot_t.days * seconds_per_day
                seconds_elapsed += tot_t.seconds
                seconds_elapsed += 1e-6 * tot_t.microseconds
                minutes_elapsed = seconds_elapsed / 60
                hours_elapsed = minutes_elapsed / 60
                days_elapsed = hours_elapsed / 24

                ns_per_day = ns_sim / days_elapsed
                hours_per_ns = hours_elapsed / ns_sim
                steps_per_s = (step + 1) / seconds_elapsed
                info_str = (
                    f"MD step = {step:10d}   step time: "
                    f"{format_timedelta(step_t):22s}   Performance: "
                    f"{ns_per_day:.3f} ns/day   {hours_per_ns:.3f} hours/ns   "
                    f"{steps_per_s:.3f} steps/s"
                )
                Logger.rank0.log(logging.INFO, info_str)

        # Initial outer rRESPA velocity step
        velocities = integrate_velocity(
            velocities, field_forces / config.mass,
            config.respa_inner * config.time_step,
        )
        if charges_flag and config.coulombtype == "PIC_Spectral":
            velocities = integrate_velocity(
                velocities, elec_forces / config.mass,
                config.respa_inner * config.time_step,
            )
        if protein_flag:
            velocities = integrate_velocity(
                velocities, reconstructed_forces / config.mass,
                config.respa_inner * config.time_step,
            )

        # Inner rRESPA steps
        for inner in range(config.respa_inner):
            velocities = integrate_velocity(
                velocities,
                (bond_forces + angle_forces + dihedral_forces) / config.mass,
                config.time_step,
            )
            positions = integrate_position(
                positions, velocities, config.time_step
            )
            positions = np.mod(positions, config.box_size[None, :])

            # Update fast forces
            if molecules_flag:
                if not args.disable_bonds:
                    bond_energy_ = compute_bond_forces(
                        bond_forces, positions, config.box_size, bonds_2_atom1,
                        bonds_2_atom2, bonds_2_equilibrium, bonds_2_stength,
                    )
                if not args.disable_angle_bonds:
                    angle_energy_ = compute_angle_forces(
                        angle_forces, positions, config.box_size,
                        bonds_3_atom1, bonds_3_atom2, bonds_3_atom3,
                        bonds_3_equilibrium, bonds_3_stength,
                    )
                if not args.disable_dihedrals:
                    if (
                        inner == config.respa_inner - 1
                        and not args.disable_dipole
                    ):
                        dipole_flag = 1
                    else:
                        dipole_flag = 0
                    dihedral_energy_ = compute_dihedral_forces(
                        dihedral_forces, positions, dipole_positions,
                        transfer_matrices, config.box_size, bonds_4_atom1,
                        bonds_4_atom2, bonds_4_atom3, bonds_4_atom4,
                        bonds_4_coeff, bonds_4_type, bonds_4_last, dipole_flag,
                    )

            velocities = integrate_velocity(
                velocities,
                (bond_forces + angle_forces + dihedral_forces) / config.mass,
                config.time_step,
            )

        # Update slow forces
        if not args.disable_field:
            layouts = [
                pm.decompose(positions[types == t]) for t in range(config.n_types)  # noqa: E501
            ]
            update_field(
                phi, layouts, force_on_grid, hamiltonian, pm, positions, types,
                config, v_ext, phi_fourier, v_ext_fourier,
            )
            compute_field_force(
                layouts, positions, force_on_grid, field_forces, types,
                config.n_types,
            )

            if charges_flag and config.coulombtype == "PIC_Spectral":
                layout_q = pm.decompose(positions)
                update_field_force_q(
                    charges, phi_q, phi_q_fourier, elec_field_fourier,
                    elec_field, elec_forces, layout_q, pm, positions, config,
                )
                field_q_energy = compute_field_energy_q(
                    config, phi_q_fourier, elec_energy_field, field_q_energy,
                    comm=comm,
                )

            if protein_flag and not args.disable_dipole:
                dipole_positions = np.reshape(
                    dipole_positions, (4 * n_tors, 3)
                )
                dipole_forces = np.reshape(
                    dipole_forces, (4 * n_tors, 3)
                )

                layout_dipoles = pm.decompose(dipole_positions)
                update_field_force_q(
                    dipole_charges, phi_dipoles, phi_dipoles_fourier,
                    dipoles_field_fourier, dipoles_field, dipole_forces,
                    layout_dipoles, pm, dipole_positions, config,
                )

                dipole_positions = np.reshape(dipole_positions, (n_tors, 4, 3))
                dipole_forces = np.reshape(dipole_forces, (n_tors, 4, 3))

                dipole_positions = np.asfortranarray(dipole_positions)
                dipole_forces_redistribution(
                    reconstructed_forces, dipole_forces, transfer_matrices,
                    bonds_4_atom1, bonds_4_atom2, bonds_4_atom3, bonds_4_atom4,
                    bonds_4_type, bonds_4_last,
                )

        # Second rRESPA velocity step
        velocities = integrate_velocity(
            velocities, field_forces / config.mass,
            config.respa_inner * config.time_step,
        )

        if charges_flag and config.coulombtype == "PIC_Spectral":
            velocities = integrate_velocity(
                velocities, elec_forces / config.mass,
                config.respa_inner * config.time_step,
            )
        if protein_flag and not args.disable_dipole:
            velocities = integrate_velocity(
                velocities, reconstructed_forces / config.mass,
                config.respa_inner * config.time_step,
            )

        # Only compute and keep the molecular bond energy from the last rRESPA
        # inner step
        if molecules_flag:
            if not args.disable_bonds:
                bond_energy = comm.allreduce(bond_energy_, MPI.SUM)
            if not args.disable_angle_bonds:
                angle_energy = comm.allreduce(angle_energy_, MPI.SUM)
            if not args.disable_dihedrals:
                dihedral_energy = comm.allreduce(dihedral_energy_, MPI.SUM)

        if step != 0 and config.domain_decomposition:
            if np.mod(step, config.domain_decomposition) == 0:
                positions = np.ascontiguousarray(positions)
                bond_forces = np.ascontiguousarray(bond_forces)
                angle_forces = np.ascontiguousarray(angle_forces)
                dihedral_forces = np.ascontiguousarray(dihedral_forces)

                args_in = [
                    velocities, indices, bond_forces, angle_forces,
                    dihedral_forces, reconstructed_forces, field_forces, names,
                    types,
                ]

                if charges_flag:
                    args_in.append(charges)
                    args_in.append(elec_forces)

                dd = domain_decomposition(  # noqa: F841
                    positions, pm, *tuple(args_in),
                    molecules=molecules if molecules_flag else None,
                    bonds=bonds if molecules_flag else None,
                    verbose=args.verbose, comm=comm,
                )
                exec(_cmd_receive_dd)

                positions = np.asfortranarray(positions)
                bond_forces = np.asfortranarray(bond_forces)
                angle_forces = np.asfortranarray(angle_forces)
                dihedral_forces = np.asfortranarray(dihedral_forces)

                layouts = [
                    pm.decompose(positions[types == t]) for t in range(config.n_types)  # noqa: E501
                ]
                if molecules_flag:
                    bonds_prep = prepare_bonds(
                        molecules, names, bonds, indices, config
                    )
                    (
                        # two-particle bonds
                        bonds_2_atom1, bonds_2_atom2, bonds_2_equilibrium,
                        bonds_2_stength,
                        # three-particle bonds
                        bonds_3_atom1, bonds_3_atom2, bonds_3_atom3,
                        bonds_3_equilibrium, bonds_3_stength,
                        # four-particle bonds
                        bonds_4_atom1, bonds_4_atom2, bonds_4_atom3,
                        bonds_4_atom4, bonds_4_coeff, bonds_4_type,
                        bonds_4_last,
                    ) = bonds_prep

                    bonds_4_coeff = np.asfortranarray(bonds_4_coeff)

                    # Reinitialize dipoles so each rank has the right amount
                    if protein_flag and not args.disable_dipole:
                        # Each rank will have different n_tors, we don't need
                        # to domain decompose dipoles
                        n_tors = len(bonds_4_atom1)
                        dipole_positions = np.zeros(
                            (n_tors, 4, 3), dtype=dtype
                        )

                        # 4 because we need to take into account the last angle
                        # in the molecule
                        dipole_charges = np.array(
                            [
                                2 * [0.25, -0.25]
                                if (bonds_4_type[i], bonds_4_last[i]) == (1, 1)
                                else [0.25, -0.25, 0.0, 0.0]
                                if bonds_4_type[i] == 1
                                else 2 * [0.0, 0.0]
                                for i in range(n_tors)
                            ],
                            dtype=dtype,
                        ).flatten()
                        dipole_forces = np.zeros_like(dipole_positions)
                        transfer_matrices = np.zeros(
                            shape=(n_tors, 6, 3, 3), dtype=dtype
                        )
                        dipole_positions = np.asfortranarray(dipole_positions)
                        transfer_matrices = np.asfortranarray(transfer_matrices)  # noqa: E501

        for t in range(config.n_types):
            if args.verbose > 2:
                exchange_cost = layouts[t].get_exchange_cost()
                Logger.all_ranks.log(
                    logging.INFO,
                    (
                        f"(GHOSTS: Total number of particles of type "
                        f"{config.type_to_name_map[t]} to be "
                        f"exchanged = {exchange_cost[rank]}"
                    ),
                )

        # Thermostat
        if config.target_temperature:
            csvr_thermostat(velocities, names, config, comm=comm)

        # Remove total linear momentum
        if config.cancel_com_momentum:
            if np.mod(step, config.cancel_com_momentum) == 0:
                velocities = cancel_com_momentum(velocities, config, comm=comm)

        # Print trajectory
        if config.n_print > 0:
            if np.mod(step, config.n_print) == 0 and step != 0:
                frame = step // config.n_print
                if not args.disable_field:
                    (
                        field_energy, kinetic_energy,
                    ) = compute_field_and_kinetic_energy(
                        phi, velocities, hamiltonian, positions, types, v_ext,
                        config, layouts, comm=comm,
                    )

                    if charges_flag and config.coulombtype == "PIC_Spectral":
                        field_q_energy = compute_field_energy_q(
                            config, phi_q_fourier, elec_energy_field,
                            field_q_energy, comm=comm,
                        )
                else:
                    kinetic_energy = comm.allreduce(
                        0.5 * config.mass * np.sum(velocities ** 2)
                    )
                temperature = (
                    (2 / 3) * kinetic_energy
                    / (config.gas_constant * config.n_particles)
                )
                if args.disable_field:
                    field_energy = 0.0

                forces_out = (
                    field_forces + bond_forces + angle_forces
                    + dihedral_forces + reconstructed_forces
                )
                store_data(
                    out_dataset, step, frame, indices, positions, velocities,
                    forces_out, config.box_size, temperature, kinetic_energy,
                    bond_energy, angle_energy, dihedral_energy, field_energy,
                    field_q_energy, config.respa_inner * config.time_step,
                    config, velocity_out=args.velocity_output,
                    force_out=args.force_output, charge_out=charges_flag,
                    dump_per_particle=args.dump_per_particle, comm=comm,
                )
                if np.mod(step, config.n_print * config.n_flush) == 0:
                    out_dataset.flush()
        last_step_time = current_step_time

    # End simulation
    if rank == 0:
        end_time = datetime.datetime.now()
        sim_time = end_time - start_time
        setup_time = loop_start_time - start_time
        loop_time = end_time - loop_start_time
        Logger.rank0.log(
            logging.INFO,
            (
                f"Elapsed time: {format_timedelta(sim_time)}   "
                f"Setup time: {format_timedelta(setup_time)}   "
                f"MD loop time: {format_timedelta(loop_time)}"
            ),
        )

    if config.n_print > 0 and np.mod(config.n_steps - 1, config.n_print) != 0:
        if not args.disable_field:
            update_field(
                phi, layouts, force_on_grid, hamiltonian, pm, positions, types,
                config, v_ext, phi_fourier, v_ext_fourier,
                compute_potential=True,
            )
            field_energy, kinetic_energy = compute_field_and_kinetic_energy(
                phi, velocities, hamiltonian, positions, types, v_ext, config,
                layouts, comm=comm,
            )

            if charges_flag and config.coulombtype == "PIC_Spectral":
                layout_q = pm.decompose(positions)
                update_field_force_q(
                    charges, phi_q, phi_q_fourier, elec_field_fourier,
                    elec_field, elec_forces, layout_q, pm, positions, config,
                )

                field_q_energy = compute_field_energy_q(
                    config, phi_q_fourier, elec_energy_field, field_q_energy,
                    comm=comm,
                )

        else:
            kinetic_energy = comm.allreduce(
                0.5 * config.mass * np.sum(velocities**2)
            )
        frame = (step + 1) // config.n_print
        temperature = (
            (2 / 3) * kinetic_energy
            / (config.gas_constant * config.n_particles)
        )
        if args.disable_field:
            field_energy = 0.0
        forces_out = (
            field_forces + bond_forces + angle_forces + dihedral_forces
            + reconstructed_forces
        )
        store_data(
            out_dataset, step, frame, indices, positions, velocities,
            forces_out, config.box_size, temperature, kinetic_energy,
            bond_energy, angle_energy, dihedral_energy, field_energy,
            field_q_energy, config.respa_inner * config.time_step, config,
            velocity_out=args.velocity_output, force_out=args.force_output,
            charge_out=charges_flag, dump_per_particle=args.dump_per_particle,
            comm=comm,
        )

    out_dataset.close_file()
